/*
 * GPU test for Spacemit
 * Copyright (C) 2024 Spacemit Co., Ltd.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <utils_opengles.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <gbm.h>
#include <libdrm/drm_fourcc.h>
#include <GLES2/gl2ext.h>

#include <sys/ioctl.h>
#include <linux/dma-heap.h>
#include <drm/drm_fourcc.h>
#include <errno.h>

bool ready_to_render = false;
pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cv = PTHREAD_COND_INITIALIZER;

#define USE_GBM_BUF 1
#define USE_DMA_BUF 0

static void signal_int(int signum)
{
	global_state.running = 0;
}

typedef struct
{
	// Handle to a program object
	GLuint programObject;
	// Uniform locations
	GLint mvpLoc;
	// Vertex data
	GLfloat *vertices;
	GLuint *indices;
	int num_indices[2];
	// Rotation angle
	GLfloat angle;
	uint32_t rotate_benchmark_time;
	// MVP matrix
	ESMatrix mvpMatrix;
	ESMatrix perspective;
	ESMatrix modelview;
	// vertex array, buffer
	GLuint vao, vbo, ebo, lineEbo;
	GLuint textureID;

	GLuint *textures;
	int current_texture_index;

} UserData;

//
// Initialize the shader and program object
//
int Init(struct Window *window)
{
	UserData *userData = (UserData *)window->userData;

	char vertShaderSrc[] = {
		"#version 300 es\n"
		"layout(location = 0) in vec3 vPosition;\n"
		"layout(location = 1) in vec2 aTexCoord;\n"
		"layout(location = 2) in vec4 aColor;\n"
		"out vec2 vTextureCoord;\n"
		"out vec4 vColor;\n"
		"uniform mat4 u_mvpMatrix;\n"
		"void main() {\n"
		"    gl_Position = u_mvpMatrix * vec4(vPosition, 1.0f);\n"
		"    vTextureCoord = vec2(aTexCoord.x, 1.0 - aTexCoord.y);\n"
		"    vColor = aColor.argb;\n"
		"}\n"};

	char fragShaderSrc[] = {
		"#version 300 es\n"
		"precision mediump float;\n"
		"uniform sampler2D uSampler;\n"
		"uniform bool useTexture;\n"
		"in vec2 vTextureCoord;\n"
		"in vec4 vColor;\n"
		"out vec4 outColor;\n"
		"void main() {\n"
		"	if(useTexture){\n"
		"    	outColor = texture(uSampler, vTextureCoord).bgra;\n"
		"	} else{\n"
		"		outColor = vColor;\n"
		"	}"
		"}\n"};

	userData->programObject = load_program(vertShaderSrc, fragShaderSrc);

	// Define the vertex positions of the cube
	GLfloat vertices[] = {
		// vertex		// texture
		0.5f, -0.5f, 0.5f, 1.0f, 0.0f,	 // point 0
		-0.5f, -0.5f, 0.5f, 0.0f, 0.0f,	 // point 1
		-0.5f, 0.5f, 0.5f, 0.0f, 1.0f,	 // point 2
		0.5f, 0.5f, 0.5f, 1.0f, 1.0f,	 // point 3
		0.5f, -0.5f, -0.5f, 0.0f, 0.0f,	 // point 4
		-0.5f, -0.5f, -0.5f, 1.0f, 0.0f, // point 5
		-0.5f, 0.5f, -0.5f, 1.0f, 1.0f,	 // point 6
		0.5f, 0.5f, -0.5f, 0.0f, 1.0f,	 // point 7

		0.5f, -0.5f, 0.5f, 1.0f, 1.0f,	 // point 0->8
		-0.5f, -0.5f, 0.5f, 0.0f, 1.0f,	 // point 1->9
		-0.5f, 0.5f, 0.5f, 0.0f, 0.0f,	 // point 2->10
		0.5f, 0.5f, 0.5f, 1.0f, 0.0f,	 // point 3->11
		0.5f, -0.5f, -0.5f, 1.0f, 0.0f,	 // point 4->12
		-0.5f, -0.5f, -0.5f, 0.0f, 0.0f, // point 5->13
		-0.5f, 0.5f, -0.5f, 0.0f, 1.0f,	 // point 6->14
		0.5f, 0.5f, -0.5f, 1.0f, 1.0f,	 // point 7->15
	};

	// Define the vertex index of a cube for drawing solid faces
	GLuint indices[] = {
		0, 1, 2, 2, 3, 0, // front
		1, 5, 6, 6, 2, 1, // left
		4, 5, 6, 6, 7, 4, // back
		0, 4, 7, 7, 3, 0, // right

		8, 12, 13, 13, 9, 8,	// bottom
		11, 10, 15, 15, 14, 10, // top
	};

	// Define the wireframe index of the cube
	GLuint lineIndices[] = {
		0, 1, 1, 2, 2, 3, 3, 0, // front
		4, 5, 5, 6, 6, 7, 7, 4, // back
		0, 4, 1, 5, 2, 6, 3, 7, // sides
	};

	userData->vertices = (GLfloat *)malloc(sizeof(vertices));
	memcpy(userData->vertices, vertices, sizeof(vertices));

	userData->num_indices[0] = sizeof(indices) / sizeof(indices[0]);
	userData->num_indices[1] = sizeof(lineIndices) / sizeof(lineIndices[0]);

	// Bind Vertex Array Object (VAO) and Vertex Buffer Object (VBO), and send vertex data to GPU
	glGenVertexArrays(1, &userData->vao);
	glBindVertexArray(userData->vao);

	glGenBuffers(1, &userData->vbo);
	glBindBuffer(GL_ARRAY_BUFFER, userData->vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), userData->vertices, GL_STATIC_DRAW);

	// Copy the index array into an index buffer for OpenGL to use
	glGenBuffers(1, &userData->ebo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, userData->ebo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	// Create wireframe EBO
	glGenBuffers(1, &userData->lineEbo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, userData->lineEbo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(lineIndices), lineIndices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	// Set vertex pointer attribute
	GLuint positionAttrib = glGetAttribLocation(userData->programObject, "vPosition");
	glVertexAttribPointer(positionAttrib, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), (void *)0);
	glEnableVertexAttribArray(positionAttrib);

	// Texture position attribute
	GLuint textureAttrib = glGetAttribLocation(userData->programObject, "aTexCoord");
	glVertexAttribPointer(textureAttrib, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), (void *)(3 * sizeof(GLfloat)));
	glEnableVertexAttribArray(textureAttrib);

	// Compute the window aspect ratio
	float aspect;
	aspect = (GLfloat)window->geometry.width / (GLfloat)window->geometry.height;
	// Generate a perspective matrix with a 60 degree FOV
	esMatrixLoadIdentity(&userData->perspective);
	esPerspective(&userData->perspective, 60.0f, aspect, 1.0f, 15.0f);

	// Generate a model view matrix to rotate/translate the cube
	esMatrixLoadIdentity(&userData->modelview);

	// Translate away from the viewer
	esTranslate(&userData->modelview, 0.0, 0.0, -2.0);

	// Starting rotation angle for the cube
	userData->angle = 0.0f;
	// rotate the matrix
	esRotate(&userData->modelview, userData->angle, 0.5, 1.0, 0.0);

	// Generate a mvp matrix
	esMatrixMultiply(&userData->mvpMatrix, &userData->modelview, &userData->perspective);

	return 1;
}

void Update(struct Window *window)
{
	struct Display *display = window->display;
	UserData *userData = (UserData *)window->userData;
	// Rotation angle update interval
	static uint8_t update_interval;
	static const uint8_t initial_update_interval = 3;
	static const uint8_t subsequent_update_interval = 3;
	static uint8_t call_count = 0;
	struct timeval tv;

	if (gettimeofday(&tv, NULL) == -1)
	{
		printf("get time of day faild.");
	}

	uint32_t time = tv.tv_sec * 1000 + tv.tv_usec / 1000;
	// Init rotate_benchmark_time and update_interval
	if (call_count == 0)
	{
		userData->rotate_benchmark_time = time;
		update_interval = initial_update_interval;
	}
	// Update rotate_benchmark_time and angle
	if (time - userData->rotate_benchmark_time > (update_interval * 1000))
	{
		update_interval = subsequent_update_interval;
		userData->rotate_benchmark_time = time;
		userData->angle += 3.0f;
		if (userData->angle > 6.0f)
		{
			userData->angle = 1.0f;
		}
	}

	// Update the model view projection matrix
	esRotate(&userData->modelview, userData->angle, 0.5, 1.0, 0.0);
	esMatrixMultiply(&userData->mvpMatrix, &userData->modelview, &userData->perspective);

	call_count = 1;
}

void Draw(struct Window *window, struct wl_callback *callback, uint32_t time)
{

	struct Display *display = window->display;

	UserData *userData = (UserData *)window->userData;

	static const uint32_t benchmark_interval = 5;
	struct wl_region *region;
	struct timeval tv;

	if (window->callback != callback)
	{
		log_error("window->callback != callback");
	}
	window->callback = NULL;

	if (callback)
		wl_callback_destroy(callback);

	if (gettimeofday(&tv, NULL) == -1)
	{
		printf("get time of day faild.");
	}

	time = tv.tv_sec * 1000 + tv.tv_usec / 1000;
	if (window->frames == 0)
		window->benchmark_time = time;
	if (time - window->benchmark_time > (benchmark_interval * 1000))
	{
		printf("%d frames in %d seconds: %f fps\n",
			   window->frames,
			   benchmark_interval,
			   (float)window->frames / benchmark_interval);
		window->benchmark_time = time;
		window->frames = 0;
	}

	// Update the model view projection matrix
	Update(window);

	// Set the viewport
	glViewport(0, 0, window->geometry.width, window->geometry.height);
	// Clear the color buffer
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glDisable(GL_BLEND);

	// Use the program object
	glUseProgram(userData->programObject);

	// Load the MVP matrix
	glUniformMatrix4fv(glGetUniformLocation(userData->programObject, "u_mvpMatrix"), 1, GL_FALSE, (GLfloat *)&userData->mvpMatrix.m[0][0]);

	glBindVertexArray(userData->vao);

	GLenum error = glGetError();
	if (error != GL_NO_ERROR)
	{
		printf("OpenGL Error: %d\n", error);
	}

	// Rendering Cube
	glUniform1i(glGetUniformLocation(userData->programObject, "useTexture"), GL_TRUE);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, userData->textures[userData->current_texture_index]); // Bind the correct texture
	glUniform1i(glGetUniformLocation(userData->programObject, "uSampler"), 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, userData->ebo);
	glDrawElements(GL_TRIANGLES, userData->num_indices[0], GL_UNSIGNED_INT, 0);

	// Rendering wireframe
	glUniform1i(glGetUniformLocation(userData->programObject, "useTexture"), GL_FALSE);
	glBindTexture(GL_TEXTURE_2D, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, userData->lineEbo);
	glLineWidth(5.0f); // Set line width
	const GLfloat wireframeColor[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glVertexAttrib4fv(2, wireframeColor); // Set color
	glDrawElements(GL_LINES, userData->num_indices[1], GL_UNSIGNED_INT, 0);

	// Swap buffer
	eglSwapBuffers(display->egl.dpy, window->egl_surface);

	window->frames++;
}

// Cleanup
void Shutdown(void *data)
{
	struct Window *window = data;
	UserData *userData = window->userData;

	if (userData->vbo != 0)
	{
		glDeleteBuffers(1, &userData->vbo);
		userData->vbo = 0;
	}

	if (userData->vao != 0)
	{
		glDeleteBuffers(1, &userData->vao);
		userData->vao = 0;
	}

	if (userData->ebo != 0)
	{
		glDeleteBuffers(1, &userData->ebo);
		userData->ebo = 0;
	}

	if (userData->vertices != NULL)
	{
		free(userData->vertices);
	}

	if (userData->indices != NULL)
	{
		free(userData->indices);
	}
	// Delete program object
	glDeleteProgram(userData->programObject);
}

#if USE_GBM_BUF
#define MAX_BUFFER_PLANES 4
struct Buffer
{
	struct Display *display;
	struct wl_buffer *buffer;
	int busy;

	struct gbm_bo *bo;
	void *map;
	void *map_data;

	int width;
	int height;
	int format;
	uint64_t modifier;
	int plane_count;
	int dmabuf_fds[MAX_BUFFER_PLANES];
	uint32_t strides[MAX_BUFFER_PLANES];
	uint32_t offsets[MAX_BUFFER_PLANES];

	EGLImageKHR egl_image;
	GLuint gl_texture;
	GLuint gl_fbo;

	struct zwp_linux_buffer_release_v1 *buffer_release;
	/* The buffer owns the release_fence_fd, until it passes ownership
	 * to it to EGL (see wait_for_buffer_release_fence). */
	int release_fence_fd;
};

static bool display_set_up_gbm(struct Display *display, const char *drm_render_node)
{
	fprintf(stdout, "display_set_up_gbm...\n");
	display->gbm.drm_fd = open(drm_render_node, O_RDWR);
	if (display->gbm.drm_fd < 0)
	{
		fprintf(stderr, "Failed to open drm render node %s\n",
				drm_render_node);
		return false;
	}

	display->gbm.device = gbm_create_device(display->gbm.drm_fd);
	if (display->gbm.device == NULL)
	{
		fprintf(stderr, "Failed to create gbm device\n");
		return false;
	}

	return true;
}

// Create GBM buffer
bool create_gbm_buffer(struct Display *display, struct Buffer *buffer, int width, int height)
{
	buffer->format = GBM_FORMAT_ARGB8888;
	// fprintf(stdout, "buffer format: %x\n", buffer->format);
	buffer->bo = gbm_bo_create(display->gbm.device, width, height, buffer->format, GBM_BO_USE_RENDERING);
	if (!buffer->bo)
	{
		fprintf(stderr, "Failed to create GBM buffer\n");
		return false;
	}

	// Specify the area to be mapped
	uint32_t x = 0;
	uint32_t y = 0;
	uint32_t flags = GBM_BO_TRANSFER_WRITE;
	uint32_t stride;

	void *map;
	void *map_data;
	map = gbm_bo_map(buffer->bo, x, y, width, height, flags, &stride, &map_data);
	if (!map)
	{
		fprintf(stderr, "Failed to map GBM buffer\n");
		gbm_bo_destroy(buffer->bo);
		return false;
	}
	// Save the mapped pointer to gbm_buf ->map
	buffer->map = map;
	buffer->strides[0] = stride;
	buffer->offsets[0] = 0;
	buffer->dmabuf_fds[0] = gbm_bo_get_fd(buffer->bo);
	buffer->width = width;
	buffer->height = height;
	buffer->map_data = map_data;

	return true;
}

bool save_buffer_to_bin(struct Buffer *buffer, const char *filename, int total_size)
{
	FILE *file = fopen(filename, "wb"); // Open file in binary write mode
	if (!file)
	{
		fprintf(stderr, "Failed to open file '%s' for writing: %s\n", filename, strerror(errno));
		return false;
	}

	size_t bytes_written = fwrite(buffer->map, 1, total_size, file); // Write data

	if (bytes_written != total_size)
	{
		fprintf(stderr, "Failed to write all data to file '%s'\n", filename);
		fclose(file);
		return false;
	}

	fclose(file);
	return true;
}

GLuint create_texture_gbm(struct Display *display, const char *drm_render_node, struct Buffer *buffer, const char *filename)
{
	if (!display_set_up_gbm(display, drm_render_node))
	{
		fprintf(stderr, "init gbm faild!");
		return -1;
	}

	int width, height, nrChannels, total_size;
	unsigned char *image_data = load_imageData(filename, &width, &height, &nrChannels, 0);
	if (!image_data)
	{
		fprintf(stderr, "Failed to load image\n");
		free_imageData(image_data);
		return EXIT_FAILURE;
	}
	total_size = width * height * nrChannels;

	create_gbm_buffer(display, buffer, width, height);

	memcpy(buffer->map, image_data, total_size);
	free_imageData(image_data);

	EGLint attribs[] = {
		EGL_WIDTH, buffer->width,
		EGL_HEIGHT, buffer->height,
		EGL_LINUX_DRM_FOURCC_EXT, buffer->format,
		EGL_DMA_BUF_PLANE0_FD_EXT, buffer->dmabuf_fds[0],
		EGL_DMA_BUF_PLANE0_OFFSET_EXT, 0,
		EGL_DMA_BUF_PLANE0_PITCH_EXT, gbm_bo_get_stride(buffer->bo),
		EGL_NONE};

	fprintf(stdout, "GBM-BUF data, width: %d, height: %d, format: %x, dmabuf_fds: %d,\n",
			buffer->width, buffer->height, buffer->format, buffer->dmabuf_fds[0]);

	PFNEGLCREATEIMAGEKHRPROC create_image;
	create_image = (void *)eglGetProcAddress("eglCreateImageKHR");
	buffer->egl_image = create_image(display->egl.dpy, EGL_NO_CONTEXT, EGL_LINUX_DMA_BUF_EXT,
									 NULL, attribs);

	if (buffer->egl_image == EGL_NO_IMAGE_KHR)
	{
		fprintf(stderr, "EGLImageKHR creation failed, eglGetError():0x%x\n", eglGetError());
		return false;
	}

	GLuint ex_texture;
	glGenTextures(1, &ex_texture);
	glBindTexture(GL_TEXTURE_2D, ex_texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	PFNGLEGLIMAGETARGETTEXTURE2DOESPROC image_target_texture_2d; // require openGL ES2.0
	image_target_texture_2d = (void *)eglGetProcAddress("glEGLImageTargetTexture2DOES");
	image_target_texture_2d(GL_TEXTURE_2D, buffer->egl_image);
	// fprintf(stdout, "debug:image_target_texture_2d: glGetError(): %d\n", glGetError());
	glBindTexture(GL_TEXTURE_2D, 0);

	return ex_texture;
}

// Define a structure to store image data
typedef struct
{
	unsigned char *data;
	int width;
	int height;
	int nrChannels;
	size_t size;
} ImageData;

// Dma buf updates thread function
void *dmabuf_update_thread(void *arg)
{
	struct Buffer *buffer = (struct Buffer *)arg;
	char *file_names[] = {"/usr/data/bianbu2.png", "/usr/data/bianbu.png"};
	int num_imgs = sizeof(file_names) / sizeof(file_names[0]);
	ImageData images[num_imgs];
	for (int i = 0; i < num_imgs; ++i)
	{
		images[i].data = load_imageData(file_names[i], &images[i].width, &images[i].height, &images[i].nrChannels, 0);
		if (images[i].data == NULL)
		{
			fprintf(stderr, "Failed to load image %d.\n", i);
			exit(EXIT_FAILURE);
		}
		images[i].size = images[i].width * images[i].height * images[i].nrChannels;
	}

	int current_image_index = 0;
	while (1)
	{
		current_image_index = (current_image_index + 1) % num_imgs;
		// Update buffer ->map content
		if (images[current_image_index].data != NULL)
		{
			pthread_mutex_lock(&mtx);
			memcpy(buffer->map, images[current_image_index].data, images[current_image_index].size);
			ready_to_render = 1;
			pthread_cond_signal(&cv);
			pthread_mutex_unlock(&mtx);
		}
		sleep(3);
	}
	return NULL;
}
#endif

#if USE_DMA_BUF
GLuint create_texture_dma(struct Display *display, const char *filename)
{
	int width, height, nrChannels;

	// Load image data
	unsigned char *image_data = load_imageData(filename, &width, &height, &nrChannels, 0);
	if (!image_data)
	{
		fprintf(stderr, "Failed to load image\n");
		free_imageData(image_data);
		return EXIT_FAILURE;
	}
	// printf("Image loaded with dimensions %dx%d and %d channels.\n", width, height, nrChannels);

	int img_size = width * height * nrChannels;
	// 1. Allocate DMA-BUF through dma-heap
	int dma_heap_fd = open("/dev/dma_heap/system", O_RDWR | O_CLOEXEC);
	if (dma_heap_fd < 0)
	{
		perror("Failed to open dma_heap");
		return -1;
	}

	struct dma_heap_allocation_data alloc_data = {
		.len = img_size,
		.fd_flags = O_RDWR | O_CLOEXEC,
		.heap_flags = 0};

	if (ioctl(dma_heap_fd, DMA_HEAP_IOCTL_ALLOC, &alloc_data) < 0)
	{
		perror("Failed to allocate DMA-BUF");
		close(dma_heap_fd);
		return -1;
	}

	fprintf(stdout, "DMA-BUF data, len: %zu, fd: %d, fd_flags: %d, heap_flags: %u\n",
			alloc_data.len, alloc_data.fd, alloc_data.fd_flags, alloc_data.heap_flags);

	int dma_buf_fd = alloc_data.fd;
	close(dma_heap_fd); // No longer need dma-heap file descriptor

	// 2. Map DMA-BUF to user space
	void *buffer = mmap(NULL, img_size, PROT_READ | PROT_WRITE, MAP_SHARED, dma_buf_fd, 0);
	if (buffer == MAP_FAILED)
	{
		perror("Failed to mmap DMA-BUF");
		close(dma_buf_fd);
		return -1;
	}

	// 3. Write data to buffer
	// memset(buffer, 0xF0, img_size); // Fill in 0xF0 (ARGB8888 format)
	memcpy(buffer, image_data, img_size);

	// 4. Cancel mapping
	munmap(buffer, img_size);

	EGLint attribs[] = {
		EGL_WIDTH, width,
		EGL_HEIGHT, height,
		EGL_LINUX_DRM_FOURCC_EXT, DRM_FORMAT_ARGB8888,
		EGL_DMA_BUF_PLANE0_FD_EXT, dma_buf_fd,
		EGL_DMA_BUF_PLANE0_OFFSET_EXT, 0,
		EGL_DMA_BUF_PLANE0_PITCH_EXT, width * 4,
		EGL_NONE};

	PFNEGLCREATEIMAGEKHRPROC create_image;
	create_image = (void *)eglGetProcAddress("eglCreateImageKHR");
	EGLImageKHR egl_image = create_image(display->egl.dpy, EGL_NO_CONTEXT, EGL_LINUX_DMA_BUF_EXT,
										 NULL, attribs);

	if (egl_image == EGL_NO_IMAGE_KHR)
	{
		fprintf(stderr, "EGLImageKHR creation failed, eglGetError():0x%x\n", eglGetError());
		return false;
	}

	GLuint ex_texture;
	glGenTextures(1, &ex_texture);
	glBindTexture(GL_TEXTURE_2D, ex_texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	PFNGLEGLIMAGETARGETTEXTURE2DOESPROC image_target_texture_2d; // for openGL ES2.0
	image_target_texture_2d = (void *)eglGetProcAddress("glEGLImageTargetTexture2DOES");
	image_target_texture_2d(GL_TEXTURE_2D, egl_image);
	// fprintf(stdout, "debug:image_target_texture_2d: glGetError(): %d\n", glGetError());
	glBindTexture(GL_TEXTURE_2D, 0);
	return ex_texture;
}

// Texture update thread function
void *texture_update_thread(void *arg))
{
	UserData *userData = (UserData *)arg while (1)
	{
		// Protecting critical areas
		pthread_mutex_lock(&mtx);
		userData->current_texture_index = (userData->current_texture_index + 1) % 2;
		pthread_cond_signal(&cv); // Notify rendering thread
		pthread_mutex_unlock(&mtx);
		ready_to_render = 1;
		// Pause for a period of time to observe the effect
		sleep(3);
	}
	return NULL;
}
#endif

int main(int argc, char **argv)
{
	struct sigaction sigint;
	struct Display display = {0};
	struct Window window = {0};
	int i, ret = 0;

	window.display = &display;
	display.window = &window;
	window.geometry.width = 512;
	window.geometry.height = 512;
	window.window_size = window.geometry;
	window.buffer_size = 0;
	window.frame_sync = 1;
	window.delay = 0;

	window.userData = malloc(sizeof(UserData));

	for (i = 1; i < argc; i++)
	{
		if (strcmp("-d", argv[i]) == 0 && i + 1 < argc)
			window.delay = atoi(argv[++i]);
		else if (strcmp("-f", argv[i]) == 0)
			// Set full screen display
			window.fullscreen = 1;
		else if (strcmp("-m", argv[i]) == 0)
			// Maximization mode
			window.maximized = 1;
		else if (strcmp("-o", argv[i]) == 0)
			// Create an opaque surface
			window.opaque = 1;
		else if (strcmp("-s", argv[i]) == 0)
			//  EGL configuration using 16bpp (16 bit color mode)
			window.buffer_size = 16;
		else if (strcmp("-b", argv[i]) == 0)
			window.frame_sync = 0;
		else if (strcmp("-h", argv[i]) == 0)
			usage(EXIT_SUCCESS);
		else
			usage(EXIT_FAILURE);
	}

	if (create_window(&window, &display, ret) == -1)
	{
		fprintf(stderr, "create window faild!");
		destroy_window(&window, &display);
		Shutdown(&window);
		return -1;
	}
	fprintf(stderr, "create window success.\n");

	if (!Init(&window))
	{
		fprintf(stderr, "init openGL faild!");
		return -1;
	}
	fprintf(stderr, "init openGL success.\n");
	// Handling SIGINT signals (usually interrupt signals triggered by pressing Ctrl+C)
	sigint.sa_handler = signal_int;
	sigemptyset(&sigint.sa_mask);
	sigint.sa_flags = SA_RESETHAND;
	sigaction(SIGINT, &sigint, NULL);

	char *file_names[2] = {"/usr/data/bianbu.png", "/usr/data/bianbu2.png"};
	UserData *userData = window.userData;
	userData->textures = malloc(2 * sizeof(GLuint)); // Allocate space for 2 textures
	userData->current_texture_index = 0;

	pthread_t buf_tid;
	fprintf(stdout, "create_texture via %s...\n", USE_GBM_BUF ? "GBM BUF" : "DMA BUF");

#if USE_GBM_BUF
	struct Buffer *buffer = malloc(sizeof(struct Buffer));
	buffer->display = &display;
	userData->textures[0] = create_texture_gbm(&display, "/dev/dri/card1", buffer, file_names[0]);
	// save_buffer_to_bin(buffer, "./test-buffer.bin", total_size);

	// Create dma buf update thread
	if (pthread_create(&buf_tid, NULL, &dmabuf_update_thread, buffer) != 0)
	{
		perror("Failed to create texture update thread");
		return EXIT_FAILURE;
	}
#endif

#if USE_DMA_BUF
	// Create texture update thread
	if (pthread_create(&buf_tid, NULL, &texture_update_thread, userData) != 0)
	{
		perror("Failed to create texture update thread");
		return EXIT_FAILURE;
	}
	userData->textures[0] = create_texture_dma(&display, file_names[0]);
	userData->textures[1] = create_texture_dma(&display, file_names[1]);

#endif

	while (global_state.running && ret != -1)
	{
		ret = wl_display_dispatch_pending(display.display);
		Draw(&window, NULL, 0);
	}

	if (pthread_join(buf_tid, NULL) != 0)
	{
		fprintf(stderr, "Failed to join texture update thread");
		return EXIT_FAILURE;
	}

	fprintf(stdout, "simple-egl exiting\n");

#if USE_GBM_BUF
	free(buffer);
#endif

	pthread_mutex_destroy(&mtx);
	pthread_cond_destroy(&cv);
	destroy_window(&window, &display);
	Shutdown(&window);

	return 0;
}
