/*
 * GPU test for Spacemit
 * Copyright (C) 2024 Spacemit Co., Ltd.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <utils_opengles.h>

static void signal_int(int signum)
{
	global_state.running = 0;
}

typedef struct
{
	// Handle to a program object
	GLuint programObject;
	// Vertex data
	GLfloat *vertices;
	GLuint *indices;
	// vertex array, buffer
	GLuint vao, vbo;

} UserData;

//
// Initialize the shader and program object
//
int Init(struct Window *window)
{
	UserData *userData = (UserData *)window->userData;

	char vertShaderSrc[] = {
		"#version 300 es\n"
		"layout(location = 0) in vec4 vPosition;\n"
		"void main() {\n"
		"    gl_Position = vPosition;\n"
		"}"};

	char fragShaderSrc[] = {
		"#version 300 es\n"
		"precision mediump float;\n"
		"out vec4 outColor;\n"
		"void main() {\n"
		"    outColor = vec4(0.0, 0.0, 1.0, 1.0);\n"
		"}"};

	char *pszVertShader = (char *)vertShaderSrc;
	char *pszFragShader = (char *)fragShaderSrc;

	GLuint textureVertShader = 0; // Handle used to store Vertex and Fragment shader
	GLuint textureFragShader = 0;
	GLuint textureProgramObject = 0; // Used to store the handle of the created program

	// Create Vertex Shader
	textureVertShader = load_shader(pszVertShader, GL_VERTEX_SHADER);
	// Create Fragment Shader
	textureFragShader = load_shader(pszFragShader, GL_FRAGMENT_SHADER);
	// Create rogramObject
	textureProgramObject = glCreateProgram();
	// Bind Vertex and Fragment Shader
	glAttachShader(textureProgramObject, textureVertShader);
	glAttachShader(textureProgramObject, textureFragShader);
	// Link program
	glLinkProgram(textureProgramObject);
	checkProgramLinking(textureProgramObject);
	userData->programObject = textureProgramObject;

	// data
	float vertices[] =
		{
			-0.5f, -0.5f, 0.0f, // point 1
			0.5f, -0.5f, 0.0f,	// point 2
			0.0f, 0.5f, 0.0f,	// point 3
		};

	userData->vertices = (GLfloat *)malloc(sizeof(vertices));
	memcpy(userData->vertices, vertices, sizeof(vertices));

	glGenVertexArrays(1, &userData->vao);
	glBindVertexArray(userData->vao);

	glGenBuffers(1, &userData->vbo);
	glBindBuffer(GL_ARRAY_BUFFER, userData->vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), userData->vertices, GL_STATIC_DRAW);

	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (void *)0);
	glEnableVertexAttribArray(0);

	return 1;
}

void Draw(struct Window *window, struct wl_callback *callback, uint32_t time)
{

	struct Display *display = window->display;

	UserData *userData = (UserData *)window->userData;

	static const uint32_t benchmark_interval = 5;
	struct wl_region *region;
	struct timeval tv;

	if (window->callback != callback)
	{
		log_error("window->callback != callback");
	}
	window->callback = NULL;

	if (callback)
		wl_callback_destroy(callback);

	if (gettimeofday(&tv, NULL) == -1)
	{
		printf("get time of day faild.");
	}

	time = tv.tv_sec * 1000 + tv.tv_usec / 1000;

	if (window->frames == 0)
		window->benchmark_time = time;
	if (time - window->benchmark_time > (benchmark_interval * 1000))
	{
		printf("%d frames in %d seconds: %f fps\n",
			   window->frames,
			   benchmark_interval,
			   (float)window->frames / benchmark_interval);
		window->benchmark_time = time;
		window->frames = 0;
	}

	// Set the viewport
	glViewport(0, 0, window->geometry.width, window->geometry.height);
	// Clear the color buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// Bind program object
	glUseProgram(userData->programObject);
	// Bind vao
	glBindVertexArray(userData->vao);
	glDrawArrays(GL_TRIANGLES, 0, 3);
	// Swap buffer
	eglSwapBuffers(display->egl.dpy, window->egl_surface);

	window->frames++;
}

// Cleanup
void Shutdown(void *data)
{
	struct Window *window = data;
	UserData *userData = window->userData;

	if (userData->vbo != 0)
	{
		glDeleteBuffers(1, &userData->vbo);
		userData->vbo = 0;
	}

	if (userData->vao != 0)
	{
		glDeleteBuffers(1, &userData->vao);
		userData->vao = 0;
	}

	if (userData->vertices != NULL)
	{
		free(userData->vertices);
	}

	if (userData->indices != NULL)
	{
		free(userData->indices);
	}
	// Delete program object
	glDeleteProgram(userData->programObject);
}

int main(int argc, char **argv)
{
	struct sigaction sigint;
	struct Display display = {0};
	struct Window window = {0};
	int i, ret = 0;

	window.display = &display;
	display.window = &window;
	window.geometry.width = 512;
	window.geometry.height = 512;
	window.window_size = window.geometry;
	window.buffer_size = 0;
	window.frame_sync = 1;
	window.delay = 0;

	window.userData = malloc(sizeof(UserData));

	for (i = 1; i < argc; i++)
	{
		if (strcmp("-d", argv[i]) == 0 && i + 1 < argc)
			window.delay = atoi(argv[++i]);
		else if (strcmp("-f", argv[i]) == 0)
			// Set full screen display
			window.fullscreen = 1;
		else if (strcmp("-m", argv[i]) == 0)
			// Maximization mode
			window.maximized = 1;
		else if (strcmp("-o", argv[i]) == 0)
			// Create an opaque surface
			window.opaque = 1;
		else if (strcmp("-s", argv[i]) == 0)
			//  EGL configuration using 16bpp (16 bit color mode)
			window.buffer_size = 16;
		else if (strcmp("-b", argv[i]) == 0)
			window.frame_sync = 0;
		else if (strcmp("-h", argv[i]) == 0)
			usage(EXIT_SUCCESS);
		else
			usage(EXIT_FAILURE);
	}

	if (create_window(&window, &display, ret) == -1)
	{
		fprintf(stderr, "Create window faild!");
		destroy_window(&window, &display);
		Shutdown(&window);
		return -1;
	}
	fprintf(stderr, "Create window success.\n");

	if (!Init(&window))
	{
		fprintf(stderr, "Init openGL faild!");
		return -1;
	}
	fprintf(stderr, "Init openGL success.\n");

	sigint.sa_handler = signal_int;
	sigemptyset(&sigint.sa_mask);
	sigint.sa_flags = SA_RESETHAND;
	sigaction(SIGINT, &sigint, NULL);

	while (global_state.running && ret != -1)
	{
		ret = wl_display_dispatch_pending(display.display);
		Draw(&window, NULL, 0);
	}

	fprintf(stderr, "Simple-egl exiting\n");
	destroy_window(&window, &display);
	Shutdown(&window);

	return 0;
}
